use std::sync::Arc;

use fred::prelude::*;

use signaling::{IceCandidate, SdpMessage, SignalingMessage};
use tokio::sync::{
    mpsc::{Receiver, Sender},
    Mutex,
};

enum RedisChannel {
    DiscoverRequest,
    DiscoverResponse(String),
    SessionOffer(String),
    SessionAnswer(String),
    SessionIceCandidate(String),
}

impl ToString for RedisChannel {
    fn to_string(&self) -> String {
        match self {
            RedisChannel::DiscoverRequest => format!("discover"),
            RedisChannel::DiscoverResponse(name) => format!("discover:{name}"),
            RedisChannel::SessionOffer(name) => format!("offer:{name}"),
            RedisChannel::SessionAnswer(name) => format!("answer:{name}"),
            RedisChannel::SessionIceCandidate(name) => format!("icecandidate:{name}"),
        }
    }
}

impl TryFrom<&str> for RedisChannel {
    type Error = String;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        if let Some((left, right)) = value.split_once(":") {
            if left == "discover" {
                return Ok(Self::DiscoverResponse(right.to_string()));
            }
            if left == "offer" {
                return Ok(Self::SessionOffer(right.to_string()));
            }
            if left == "answer" {
                return Ok(Self::SessionAnswer(right.to_string()));
            }
            if left == "icecandidate" {
                return Ok(Self::SessionIceCandidate(right.to_string()));
            }
        }
        if value == "discover" {
            return Ok(Self::DiscoverRequest);
        }
        return Err(value.to_string());
    }
}

pub async fn send_message_to_peers(
    mut messages_to_peers: Receiver<SignalingMessage>,
    messages_from_peers: Sender<SignalingMessage>,
    redis: fred::clients::RedisClient,
    pubsub: fred::clients::SubscriberClient,
) {
    let mut name = String::default();

    let (closing_tx, closing_rx) = tokio::sync::mpsc::channel(1);
    let closing_rx = Arc::new(Mutex::new(closing_rx));

    while let Some(message) = messages_to_peers.recv().await {
        let inner_message = message.message.unwrap();
        let room: String = message.room;

        match inner_message {
            signaling::signaling_message::Message::Bootstrap(()) => { // client may need a name, bootstrap all the information here.
                if message.sender != "" {
                    name = message.sender.to_string();
                } else {
                    let mut gen = names::Generator::default();
                    name = gen.next().unwrap();
                }

                tokio::spawn(receive_message_from_peers(
                    name.clone(),
                    room,
                    messages_from_peers.clone(),
                    pubsub.clone(),
                    closing_rx.clone(),
                ));
            }
            signaling::signaling_message::Message::DiscoverRequest(()) => { // client broadcast discover to room
                let peers: usize = match redis
                    .publish(format!("chatchat:{room}:discover"), name.clone())
                    .await
                {
                    Ok(peers) => peers,
                    Err(error) => {
                        tracing::error!(?error, "failed to publish discover request");
                        break;
                    }
                };
                tracing::info!(peers, room, sender = name, "broadcasting discover");
            }
            signaling::signaling_message::Message::DiscoverResponse(()) => { // client received broadcast from others. responde to it: please connect to me. also it should prepare peer connection before send this message.
                let receiver = message.receiver.unwrap();
                let peers: usize = match redis
                    .publish(format!("chatchat:{room}:discover:{receiver}"), name.clone())
                    .await
                {
                    Ok(peers) => peers,
                    Err(error) => {
                        tracing::error!(?error, "failed to publish discover request");
                        break;
                    }
                };
                tracing::info!(peers, room, sender = name, receiver, "reply to discover");
            }
            signaling::signaling_message::Message::SessionOffer(sdp) => { // client received conenct to me message. creating offer.
                let msg = serde_json::to_string(&sdp).unwrap();
                let receiver = message.receiver.unwrap();
                let peers: usize = match redis
                    .publish(format!("chatchat:{room}:offer:{receiver}"), msg)
                    .await
                {
                    Ok(peers) => peers,
                    Err(error) => {
                        tracing::error!(?error, "failed to publish offer");
                        break;
                    }
                };

                tracing::info!(peers, room, sender = name, receiver, "offering")
            }
            signaling::signaling_message::Message::SessionAnswer(sdp) => { // answer to offer
                let msg = serde_json::to_string(&sdp).unwrap();
                let receiver = message.receiver.unwrap();
                let peers: usize = match redis
                    .publish(format!("chatchat:{room}:answer:{receiver}"), msg)
                    .await
                {
                    Ok(peers) => peers,
                    Err(error) => {
                        tracing::error!(?error, "failed to publish offer");
                        break;
                    }
                };
                tracing::info!(peers, room, sender = name, receiver, "answering")
            }
            signaling::signaling_message::Message::IceCandidate(candidate) => { // trickle ice
                let msg = serde_json::to_string(&candidate).unwrap();
                let receiver = message.receiver.unwrap();
                let peers: usize = match redis
                    .publish(format!("chatchat:{room}:icecandidate:{receiver}"), msg)
                    .await
                {
                    Ok(peers) => peers,
                    Err(error) => {
                        tracing::error!(?error, "failed to publish offer");
                        break;
                    }
                };
                tracing::info!(peers, room, sender = name, receiver, "candidate")
            }
        }
    }
    tracing::debug!(name, "stopped send to peer");
    let _ = closing_tx.send(String::from("websocket exited")).await;
}
pub async fn receive_message_from_peers(
    name: String,
    room: String,
    tx: Sender<SignalingMessage>,
    pubsub: fred::clients::SubscriberClient,
    closing_rx: Arc<Mutex<Receiver<String>>>,
) {
    let mut closing_rx = closing_rx.lock().await;
    // let mut pubsub = pubsub.lock().await;

    let discover_request_channel = format!(
        "chatchat:{room}:{}",
        RedisChannel::DiscoverRequest.to_string()
    );
    let discover_response_channel = format!(
        "chatchat:{room}:{}",
        RedisChannel::DiscoverResponse(name.clone()).to_string()
    );
    let session_offer_channel = format!(
        "chatchat:{room}:{}",
        RedisChannel::SessionOffer(name.clone()).to_string()
    );
    let session_answer_channel = format!(
        "chatchat:{room}:{}",
        RedisChannel::SessionAnswer(name.clone()).to_string()
    );
    let session_ice_channel = format!(
        "chatchat:{room}:{}",
        RedisChannel::SessionIceCandidate(name.clone()).to_string()
    );

    let channels = [
        &discover_request_channel,
        &discover_response_channel,
        &session_offer_channel,
        &session_answer_channel,
        &session_ice_channel,
    ];

    pubsub.subscribe(&channels).await.unwrap();
    tracing::info!(?channels, "subscribed");

    // let mut messages = pubsub.on_message();
    let mut messages = pubsub.message_rx();

    tracing::info!(room, name = name.clone(), "connection ready");
    let _ = tx
        .send(SignalingMessage {
            room: room.clone(),
            sender: name.clone(),
            receiver: None,
            message: Some(signaling::signaling_message::Message::Bootstrap(())),
        })
        .await;

    loop {
        tokio::select! {
            message = closing_rx.recv() => {
                tracing::warn!(message, "exit due to closing rx");
                break;
            },
            maybe_message = messages.recv() => {
                let message = match maybe_message {
                    Ok(message) => message,
                    Err(error) => {
                        tracing::warn!(?error, "failed to recv message from redis");
                        break;
                    }
                };

                let channel = match message
                    .channel.to_string()
                    .strip_prefix(&format!("chatchat:{}:", room.clone()))
                {
                    Some(channel) => channel.to_string(),
                    _ => continue,
                };
                let channel = match RedisChannel::try_from(channel.as_str()) {
                    Ok(channel) => channel,
                    Err(unrecognized) => {
                        tracing::warn!(unrecognized, "unrecognized");
                        continue;
                    }
                };
                let msg_kind = message.value.kind();
                let payload = if let fred::types::RedisValue::String(ref s) = message.value {
                    s.to_string()
                } else {
                    tracing::warn!(?msg_kind, "message is not string");
                    continue;
                };

                match channel {
                    RedisChannel::DiscoverRequest => {
                        if payload == name {
                            continue;
                        }
                        tracing::info!(room, sender = payload, name, "new peer in room");
                        let _ = tx
                            .send(SignalingMessage {
                                room: room.clone(),
                                sender: payload,
                                receiver: None,
                                message: Some(signaling::signaling_message::Message::DiscoverRequest(())),
                            })
                            .await;
                    }
                    RedisChannel::DiscoverResponse(name) => {
                        tracing::info!(room, sender = payload, name, "peer connecting");
                        let _ = tx
                            .send(SignalingMessage {
                                room: room.clone(),
                                sender: payload,
                                receiver: Some(name),
                                message: Some(signaling::signaling_message::Message::DiscoverResponse(())),
                            })
                            .await;
                    }
                    RedisChannel::SessionOffer(name) => {
                        let sdp: SdpMessage = match serde_json::from_str(&payload) {
                            Ok(sdp) => sdp,
                            _ => continue,
                        };
                        tracing::info!(room, sender = sdp.sender, name, "peer offering");
                        let _ = tx
                            .send(SignalingMessage {
                                room: room.clone(),
                                sender: sdp.sender.clone(),
                                receiver: Some(name),
                                message: Some(signaling::signaling_message::Message::SessionOffer(sdp)),
                            })
                            .await;
                    }
                    RedisChannel::SessionAnswer(name) => {
                        let sdp: SdpMessage = match serde_json::from_str(&payload) {
                            Ok(sdp) => sdp,
                            _ => continue,
                        };
                        tracing::info!(room, sender = sdp.sender, name, "peer answer");
                        let _ = tx
                            .send(SignalingMessage {
                                room: room.clone(),
                                sender: sdp.sender.clone(),
                                receiver: Some(name),
                                message: Some(signaling::signaling_message::Message::SessionAnswer(sdp)),
                            })
                            .await;
                    }
                    RedisChannel::SessionIceCandidate(name) => {
                        let candidate: IceCandidate = match serde_json::from_str(&payload) {
                            Ok(candidate) => candidate,
                            _ => continue,
                        };
                        tracing::info!(room, sender = candidate.sender, name, "peer candidate");
                        let _ = tx
                            .send(SignalingMessage {
                                room: room.clone(),
                                sender: candidate.sender.clone(),
                                receiver: Some(name.clone()),
                                message: Some(signaling::signaling_message::Message::IceCandidate(
                                    candidate,
                                )),
                            })
                            .await;
                    }
                }

            }
        }
    }

    tracing::debug!(name, "stopped recv from peer");
}
