use axum::{extract::State, response::IntoResponse};
use fred::prelude::*;

use crate::{signaling::redisexchange, types::AppState};
use axum::extract::ws::Message;
use std::sync::Arc;
use tokio::sync::RwLock;

pub async fn handle_message_from_frontend(
    mut ws: axum::extract::ws::WebSocket,
    redis: fred::clients::RedisClient,
    pubsub: fred::clients::SubscriberClient,
) {
    let (peer_tx, mut peer_rx) = tokio::sync::mpsc::channel(128);
    let (browser_tx, browser_rx) = tokio::sync::mpsc::channel(128);

    tokio::spawn(redisexchange::send_message_to_peers(
        browser_rx, peer_tx, redis, pubsub,
    ));
    loop {
        tokio::select! {
            msg_from_peer = peer_rx.recv() => {
                let msg = match msg_from_peer {
                    Some(msg) => msg,
                    None => break
                };
                let _ = ws.send(axum::extract::ws::Message::Text(serde_json::to_string(&msg).unwrap())).await;
            },
            msg_from_ws = ws.recv() => {

                let msg = match msg_from_ws {
                    Some(msg) => msg,
                    None => break
                };
                let msg = match msg {
                    Ok(msg) => msg,
                    Err(error) => {
                        tracing::error!(?error, "websocket read error");
                        break;
                    }
                };
                match msg {
                    Message::Text(text) => {
                        let msg = match serde_json::from_str(&text) {
                            Ok(msg) => msg,
                            Err(error) => {
                                tracing::warn!(?error, "failed to decode websocket");
                                continue;
                            },
                        };

                        let _ = browser_tx.send(msg).await;
                    }
                    _ => continue,
                }
            }
        }
    }
}

pub async fn handle_ws_upgrade(
    upgrade: axum::extract::WebSocketUpgrade,
    State(state): State<Arc<RwLock<AppState>>>,
) -> Result<impl IntoResponse, String> {
    let state = state.read().await;

    let builder = fred::types::Builder::from_config(state.redis.clone());
    let redis = builder.build().unwrap();
    if let Err(err) = redis.init().await {
        return Err(err.to_string());
    }

    let pubsub = builder.build_subscriber_client().unwrap();
    if let Err(err) = pubsub.init().await {
        return Err(err.to_string());
    }

    Ok(upgrade.on_upgrade(|ws| handle_message_from_frontend(ws, redis, pubsub)))
}
