#[cfg(feature = "embed-static")]
static STATIC_DIR: include_dir::Dir<'_> = include_dir::include_dir!("$CARGO_MANIFEST_DIR/static");

#[cfg(feature = "extract-static")]
pub async fn extract<P: AsRef<std::path::Path>>(to: P) -> std::io::Result<()> {
    let base_path = to.as_ref();
    let mut dirs = vec![STATIC_DIR.clone()];
    tracing::info!("extracing static assets...");

    while let Some(dir) = dirs.pop() {
        for entry in dir.entries() {
            let path = base_path.join(entry.path());

            match entry {
                DirEntry::Dir(d) => {
                    tracing::trace!(dir=?d, "directory been put into queue");
                    tokio::fs::create_dir_all(&path).await?;
                    dirs.insert(0, d.clone());
                }
                DirEntry::File(f) => {
                    tracing::trace!(?path, "file extracted");
                    tokio::fs::write(path, f.contents()).await?;
                }
            }
        }
    }

    Ok(())
}

#[cfg(feature = "serve-static")]
pub use with_service::router;
#[cfg(feature = "serve-static")]
pub use with_service::StaticFiles;

#[cfg(feature = "serve-static")]
mod with_service {
    use std::{convert::Infallible, task::Poll};

    use axum::{
        body::{Bytes, Full},
        extract::Path,
        http::{Request, StatusCode},
        response::{IntoResponse, Response},
        routing::{MethodFilter, MethodRouter},
    };
    use futures::Future;
    use include_dir::Dir;

    use super::STATIC_DIR;

    async fn head(Path(static_path): Path<String>) -> Response {
        if super::STATIC_DIR.contains(static_path) {
            (StatusCode::OK, "").into_response()
        } else {
            (StatusCode::NOT_FOUND, "").into_response()
        }
    }

    async fn get(Path(static_path): Path<String>) -> Response {
        if let Some(file) = super::STATIC_DIR.get_file(static_path) {
            (StatusCode::OK, file.contents()).into_response()
        } else {
            (StatusCode::NOT_FOUND, "").into_response()
        }
    }

    #[derive(Clone)]
    pub enum ResponseFuture {
        OpenFileFuture(&'static [u8], String),
        NotFoundFuture,
    }

    impl Future for ResponseFuture {
        type Output = Result<Response, Infallible>;

        fn poll(
            self: std::pin::Pin<&mut Self>,
            _: &mut std::task::Context<'_>,
        ) -> std::task::Poll<Self::Output> {
            match self.clone() {
                ResponseFuture::OpenFileFuture(content, mime_type) => {
                    Poll::Ready(Ok(Response::builder()
                        .status(StatusCode::OK)
                        .header("Content-Type", mime_type.clone())
                        .body(axum::body::boxed(Full::new(Bytes::copy_from_slice(
                            content,
                        ))))
                        .unwrap()))
                }
                ResponseFuture::NotFoundFuture => {
                    Poll::Ready(Ok(StatusCode::NOT_FOUND.into_response()))
                }
            }
        }
    }

    #[derive(Clone, Debug)]
    pub struct StaticFiles(Dir<'static>, String);

    impl<B> tower::Service<Request<B>> for StaticFiles
    where
        B: Send + 'static,
    {
        type Response = Response;

        type Error = Infallible;

        type Future = ResponseFuture;

        fn poll_ready(
            &mut self,
            _: &mut std::task::Context<'_>,
        ) -> std::task::Poll<Result<(), Self::Error>> {
            std::task::Poll::Ready(Ok(()))
        }

        fn call(&mut self, req: Request<B>) -> Self::Future {
            let path = if let Some(path) = req.uri().path().strip_prefix(&self.1) {
                path
            } else {
                return ResponseFuture::NotFoundFuture;
            };

            let mime_type = mime_guess::from_path(path)
                .first()
                .map_or("application/octet-stream".to_string(), |m| {
                    m.essence_str().to_string()
                });

            tracing::trace!(path, "trying to get static");
            if let Some(file) = self.0.get_file(path) {
                ResponseFuture::OpenFileFuture(file.contents(), mime_type.to_string())
            } else {
                ResponseFuture::NotFoundFuture
            }
        }
    }

    impl StaticFiles {
        pub fn new() -> Self {
            StaticFiles(STATIC_DIR.clone(), "/".to_string())
        }

        pub fn strip<S: ToString>(prefix: S) -> Self {
            StaticFiles(STATIC_DIR.clone(), prefix.to_string())
        }
    }

    pub fn router() -> MethodRouter {
        MethodRouter::new()
            .on(MethodFilter::HEAD, head)
            .on(MethodFilter::GET, get)
    }
}
